import os
import psycopg2
import json
import logging
from psycopg2.extras import RealDictCursor
from dotenv import load_dotenv

load_dotenv()
logger = logging.getLogger(__name__)

class ConfigAccessor:
    """Enable dot notation access for configuration data"""
    def __init__(self, data):
        for key, value in data.items():
            if isinstance(value, dict):
                setattr(self, key, ConfigAccessor(value))
            else:
                setattr(self, key, value)

    def __repr__(self):
        return str(self.__dict__)

def get_db_connection():
    """Create PostgreSQL connection"""
    return psycopg2.connect(
        host=os.getenv("DB_HOST"),
        database=os.getenv("DB_NAME"),
        user=os.getenv("DB_USER"),
        password=os.getenv("DB_PASSWORD"),
        port=5432,
        sslmode='require'
    )

def get_config(task_name: str) -> ConfigAccessor:
    """
    Retrieve complete configuration and return as dot-accessible object
    Usage: config = get_config('appv2')
           print(config.redcap.REDCAP_API_URL)
    """
    try:
        with get_db_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""
                    SELECT key_data FROM keys 
                    WHERE task_name = %s
                    """, (task_name,))
                result = cur.fetchone()
                
                if not result:
                    raise ValueError(f"No configuration found for {task_name}")
                
                # Convert JSONB to Python dict
                config_data = result[0]
                return ConfigAccessor(config_data)
                
    except Exception as e:
        logger.error(f"Configuration error: {str(e)}")
        raise